@@ -5,54 +5,16 @@
 #include <algorithm>
 #include <vector>
 
-static const uint32_t realsr_preproc_spv_data[] = {
-    #include "realsr_preproc.spv.hex.h"
-};
-static const uint32_t realsr_preproc_fp16s_spv_data[] = {
-    #include "realsr_preproc_fp16s.spv.hex.h"
-};
-static const uint32_t realsr_preproc_int8s_spv_data[] = {
-    #include "realsr_preproc_int8s.spv.hex.h"
-};
-static const uint32_t realsr_postproc_spv_data[] = {
-    #include "realsr_postproc.spv.hex.h"
-};
-static const uint32_t realsr_postproc_fp16s_spv_data[] = {
-    #include "realsr_postproc_fp16s.spv.hex.h"
-};
-static const uint32_t realsr_postproc_int8s_spv_data[] = {
-    #include "realsr_postproc_int8s.spv.hex.h"
-};
+#include "realsr_preproc.comp.hex.h"
+#include "realsr_postproc.comp.hex.h"
+#include "realsr_preproc_tta.comp.hex.h"
+#include "realsr_postproc_tta.comp.hex.h"
 
-static const uint32_t realsr_preproc_tta_spv_data[] = {
-    #include "realsr_preproc_tta.spv.hex.h"
-};
-static const uint32_t realsr_preproc_tta_fp16s_spv_data[] = {
-    #include "realsr_preproc_tta_fp16s.spv.hex.h"
-};
-static const uint32_t realsr_preproc_tta_int8s_spv_data[] = {
-    #include "realsr_preproc_tta_int8s.spv.hex.h"
-};
-static const uint32_t realsr_postproc_tta_spv_data[] = {
-    #include "realsr_postproc_tta.spv.hex.h"
-};
-static const uint32_t realsr_postproc_tta_fp16s_spv_data[] = {
-    #include "realsr_postproc_tta_fp16s.spv.hex.h"
-};
-static const uint32_t realsr_postproc_tta_int8s_spv_data[] = {
-    #include "realsr_postproc_tta_int8s.spv.hex.h"
-};
-
-RealSR::RealSR(int gpuid, bool _tta_mode)
+RealSR::RealSR(int gpuid, bool _tta_mode, int num_threads)
 {
-    net.opt.use_vulkan_compute = true;
-    net.opt.use_fp16_packed = true;
-    net.opt.use_fp16_storage = true;
-    net.opt.use_fp16_arithmetic = false;
-    net.opt.use_int8_storage = true;
-    net.opt.use_int8_arithmetic = false;
+    vkdev = gpuid == -1 ? 0 : ncnn::get_gpu_device(gpuid);
 
-    net.set_vulkan_device(gpuid);
+    net.opt.num_threads = num_threads;
 
     realsr_preproc = 0;
     realsr_postproc = 0;
@@ -78,6 +40,14 @@ int RealSR::load(const std::wstring& parampath, const std::wstring& modelpath)
 int RealSR::load(const std::string& parampath, const std::string& modelpath)
 #endif
 {
+    net.opt.use_vulkan_compute = vkdev ? true : false;
+    net.opt.use_fp16_packed = true;
+    net.opt.use_fp16_storage = vkdev ? true : false;
+    net.opt.use_fp16_arithmetic = false;
+    net.opt.use_int8_storage = true;
+
+    net.set_vulkan_device(vkdev);
+
 #if _WIN32
     {
         FILE* fp = _wfopen(parampath.c_str(), L"rb");
@@ -107,6 +77,7 @@ int RealSR::load(const std::string& parampath, const std::string& modelpath)
 #endif
 
     // initialize preprocess and postprocess pipeline
+    if (vkdev)
     {
         std::vector<ncnn::vk_specialization_type> specializations(1);
 #if _WIN32
@@ -115,50 +86,49 @@ int RealSR::load(const std::string& parampath, const std::string& modelpath)
         specializations[0].i = 0;
 #endif
 
-        realsr_preproc = new ncnn::Pipeline(net.vulkan_device());
-        realsr_preproc->set_optimal_local_size_xyz(32, 32, 3);
-
-        realsr_postproc = new ncnn::Pipeline(net.vulkan_device());
-        realsr_postproc->set_optimal_local_size_xyz(32, 32, 3);
-
-        if (tta_mode)
         {
-            if (net.opt.use_fp16_storage && net.opt.use_int8_storage)
-                realsr_preproc->create(realsr_preproc_tta_int8s_spv_data, sizeof(realsr_preproc_tta_int8s_spv_data), specializations);
-            else if (net.opt.use_fp16_storage)
-                realsr_preproc->create(realsr_preproc_tta_fp16s_spv_data, sizeof(realsr_preproc_tta_fp16s_spv_data), specializations);
-            else
-                realsr_preproc->create(realsr_preproc_tta_spv_data, sizeof(realsr_preproc_tta_spv_data), specializations);
+            static std::vector<uint32_t> spirv;
+            static ncnn::Mutex lock;
+            {
+                ncnn::MutexLockGuard guard(lock);
+                if (spirv.empty())
+                {
+                    if (tta_mode)
+                        compile_spirv_module(realsr_preproc_tta_comp_data, sizeof(realsr_preproc_tta_comp_data), net.opt, spirv);
+                    else
+                        compile_spirv_module(realsr_preproc_comp_data, sizeof(realsr_preproc_comp_data), net.opt, spirv);
+                }
+            }
 
-            if (net.opt.use_fp16_storage && net.opt.use_int8_storage)
-                realsr_postproc->create(realsr_postproc_tta_int8s_spv_data, sizeof(realsr_postproc_tta_int8s_spv_data), specializations);
-            else if (net.opt.use_fp16_storage)
-                realsr_postproc->create(realsr_postproc_tta_fp16s_spv_data, sizeof(realsr_postproc_tta_fp16s_spv_data), specializations);
-            else
-                realsr_postproc->create(realsr_postproc_tta_spv_data, sizeof(realsr_postproc_tta_spv_data), specializations);
+            realsr_preproc = new ncnn::Pipeline(vkdev);
+            realsr_preproc->set_optimal_local_size_xyz(8, 8, 3);
+            realsr_preproc->create(spirv.data(), spirv.size() * 4, specializations);
         }
-        else
-        {
-            if (net.opt.use_fp16_storage && net.opt.use_int8_storage)
-                realsr_preproc->create(realsr_preproc_int8s_spv_data, sizeof(realsr_preproc_int8s_spv_data), specializations);
-            else if (net.opt.use_fp16_storage)
-                realsr_preproc->create(realsr_preproc_fp16s_spv_data, sizeof(realsr_preproc_fp16s_spv_data), specializations);
-            else
-                realsr_preproc->create(realsr_preproc_spv_data, sizeof(realsr_preproc_spv_data), specializations);
 
-            if (net.opt.use_fp16_storage && net.opt.use_int8_storage)
-                realsr_postproc->create(realsr_postproc_int8s_spv_data, sizeof(realsr_postproc_int8s_spv_data), specializations);
-            else if (net.opt.use_fp16_storage)
-                realsr_postproc->create(realsr_postproc_fp16s_spv_data, sizeof(realsr_postproc_fp16s_spv_data), specializations);
-            else
-                realsr_postproc->create(realsr_postproc_spv_data, sizeof(realsr_postproc_spv_data), specializations);
+        {
+            static std::vector<uint32_t> spirv;
+            static ncnn::Mutex lock;
+            {
+                ncnn::MutexLockGuard guard(lock);
+                if (spirv.empty())
+                {
+                    if (tta_mode)
+                        compile_spirv_module(realsr_postproc_tta_comp_data, sizeof(realsr_postproc_tta_comp_data), net.opt, spirv);
+                    else
+                        compile_spirv_module(realsr_postproc_comp_data, sizeof(realsr_postproc_comp_data), net.opt, spirv);
+                }
+            }
+
+            realsr_postproc = new ncnn::Pipeline(vkdev);
+            realsr_postproc->set_optimal_local_size_xyz(8, 8, 3);
+            realsr_postproc->create(spirv.data(), spirv.size() * 4, specializations);
         }
     }
 
     // bicubic 4x for alpha channel
     {
         bicubic_4x = ncnn::create_layer("Interp");
-        bicubic_4x->vkdev = net.vulkan_device();
+        bicubic_4x->vkdev = vkdev;
 
         ncnn::ParamDict pd;
         pd.set(0, 3);// bicubic
@@ -174,6 +144,12 @@ int RealSR::load(const std::string& parampath, const std::string& modelpath)
 
 int RealSR::process(const ncnn::Mat& inimage, ncnn::Mat& outimage) const
 {
+    if (!vkdev)
+    {
+        // cpu only
+        return process_cpu(inimage, outimage);
+    }
+
     const unsigned char* pixeldata = (const unsigned char*)inimage.data;
     const int w = inimage.w;
     const int h = inimage.h;
@@ -182,8 +158,8 @@ int RealSR::process(const ncnn::Mat& inimage, ncnn::Mat& outimage) const
     const int TILE_SIZE_X = tilesize;
     const int TILE_SIZE_Y = tilesize;
 
-    ncnn::VkAllocator* blob_vkallocator = net.vulkan_device()->acquire_blob_allocator();
-    ncnn::VkAllocator* staging_vkallocator = net.vulkan_device()->acquire_staging_allocator();
+    ncnn::VkAllocator* blob_vkallocator = vkdev->acquire_blob_allocator();
+    ncnn::VkAllocator* staging_vkallocator = vkdev->acquire_staging_allocator();
 
     ncnn::Option opt = net.opt;
     opt.blob_vkallocator = blob_vkallocator;
@@ -229,7 +205,7 @@ int RealSR::process(const ncnn::Mat& inimage, ncnn::Mat& outimage) const
             }
         }
 
-        ncnn::VkCompute cmd(net.vulkan_device());
+        ncnn::VkCompute cmd(vkdev);
 
         // upload
         ncnn::VkMat in_gpu;
@@ -540,8 +516,323 @@ int RealSR::process(const ncnn::Mat& inimage, ncnn::Mat& outimage) const
         }
     }
 
-    net.vulkan_device()->reclaim_blob_allocator(blob_vkallocator);
-    net.vulkan_device()->reclaim_staging_allocator(staging_vkallocator);
+    vkdev->reclaim_blob_allocator(blob_vkallocator);
+    vkdev->reclaim_staging_allocator(staging_vkallocator);
+
+    return 0;
+}
+
+int RealSR::process_cpu(const ncnn::Mat& inimage, ncnn::Mat& outimage) const
+{
+    const unsigned char* pixeldata = (const unsigned char*)inimage.data;
+    const int w = inimage.w;
+    const int h = inimage.h;
+    const int channels = inimage.elempack;
+
+    const int TILE_SIZE_X = tilesize;
+    const int TILE_SIZE_Y = tilesize;
+
+    ncnn::Option opt = net.opt;
+
+    // each tile 100x100
+    const int xtiles = (w + TILE_SIZE_X - 1) / TILE_SIZE_X;
+    const int ytiles = (h + TILE_SIZE_Y - 1) / TILE_SIZE_Y;
+
+    for (int yi = 0; yi < ytiles; yi++)
+    {
+        const int tile_h_nopad = std::min((yi + 1) * TILE_SIZE_Y, h) - yi * TILE_SIZE_Y;
+
+        int in_tile_y0 = std::max(yi * TILE_SIZE_Y - prepadding, 0);
+        int in_tile_y1 = std::min((yi + 1) * TILE_SIZE_Y + prepadding, h);
+
+        for (int xi = 0; xi < xtiles; xi++)
+        {
+            const int tile_w_nopad = std::min((xi + 1) * TILE_SIZE_X, w) - xi * TILE_SIZE_X;
+
+            int in_tile_x0 = std::max(xi * TILE_SIZE_X - prepadding, 0);
+            int in_tile_x1 = std::min((xi + 1) * TILE_SIZE_X + prepadding, w);
+
+            // crop tile
+            ncnn::Mat in;
+            {
+                if (channels == 3)
+                {
+#if _WIN32
+                    in = ncnn::Mat::from_pixels_roi(pixeldata, ncnn::Mat::PIXEL_BGR2RGB, w, h, in_tile_x0, in_tile_y0, in_tile_x1 - in_tile_x0, in_tile_y1 - in_tile_y0);
+#else
+                    in = ncnn::Mat::from_pixels_roi(pixeldata, ncnn::Mat::PIXEL_RGB, w, h, in_tile_x0, in_tile_y0, in_tile_x1 - in_tile_x0, in_tile_y1 - in_tile_y0);
+#endif
+                }
+                if (channels == 4)
+                {
+#if _WIN32
+                    in = ncnn::Mat::from_pixels_roi(pixeldata, ncnn::Mat::PIXEL_BGRA2RGBA, w, h, in_tile_x0, in_tile_y0, in_tile_x1 - in_tile_x0, in_tile_y1 - in_tile_y0);
+#else
+                    in = ncnn::Mat::from_pixels_roi(pixeldata, ncnn::Mat::PIXEL_RGBA, w, h, in_tile_x0, in_tile_y0, in_tile_x1 - in_tile_x0, in_tile_y1 - in_tile_y0);
+#endif
+                }
+            }
+
+            ncnn::Mat out;
+
+            if (tta_mode)
+            {
+                // split alpha and preproc
+                ncnn::Mat in_tile[8];
+                ncnn::Mat in_alpha_tile;
+                {
+                    in_tile[0].create(in.w, in.h, 3);
+                    for (int q = 0; q < 3; q++)
+                    {
+                        const float* ptr = in.channel(q);
+                        float* outptr0 = in_tile[0].channel(q);
+
+                        for (int i = 0; i < in.h; i++)
+                        {
+                            for (int j = 0; j < in.w; j++)
+                            {
+                                *outptr0++ = *ptr++ * (1 / 255.f);
+                            }
+                        }
+                    }
+
+                    if (channels == 4)
+                    {
+                        in_alpha_tile = in.channel_range(3, 1).clone();
+                    }
+                }
+
+                // border padding
+                {
+                    int pad_top = std::max(prepadding - yi * TILE_SIZE_Y, 0);
+                    int pad_bottom = std::max(std::min((yi + 1) * TILE_SIZE_Y + prepadding - h, prepadding), 0);
+                    int pad_left = std::max(prepadding - xi * TILE_SIZE_X, 0);
+                    int pad_right = std::max(std::min((xi + 1) * TILE_SIZE_X + prepadding - w, prepadding), 0);
+
+                    ncnn::Mat in_tile_padded;
+                    ncnn::copy_make_border(in_tile[0], in_tile_padded, pad_top, pad_bottom, pad_left, pad_right, 2, 0.f, net.opt);
+                    in_tile[0] = in_tile_padded;
+                }
+
+                // the other 7 directions
+                {
+                    in_tile[1].create(in_tile[0].w, in_tile[0].h, 3);
+                    in_tile[2].create(in_tile[0].w, in_tile[0].h, 3);
+                    in_tile[3].create(in_tile[0].w, in_tile[0].h, 3);
+                    in_tile[4].create(in_tile[0].h, in_tile[0].w, 3);
+                    in_tile[5].create(in_tile[0].h, in_tile[0].w, 3);
+                    in_tile[6].create(in_tile[0].h, in_tile[0].w, 3);
+                    in_tile[7].create(in_tile[0].h, in_tile[0].w, 3);
+
+                    for (int q = 0; q < 3; q++)
+                    {
+                        const ncnn::Mat in_tile_0 = in_tile[0].channel(q);
+                        ncnn::Mat in_tile_1 = in_tile[1].channel(q);
+                        ncnn::Mat in_tile_2 = in_tile[2].channel(q);
+                        ncnn::Mat in_tile_3 = in_tile[3].channel(q);
+                        ncnn::Mat in_tile_4 = in_tile[4].channel(q);
+                        ncnn::Mat in_tile_5 = in_tile[5].channel(q);
+                        ncnn::Mat in_tile_6 = in_tile[6].channel(q);
+                        ncnn::Mat in_tile_7 = in_tile[7].channel(q);
+
+                        for (int i = 0; i < in_tile[0].h; i++)
+                        {
+                            const float* outptr0 = in_tile_0.row(i);
+                            float* outptr1 = in_tile_1.row(in_tile[0].h - 1 - i);
+                            float* outptr2 = in_tile_2.row(i) + in_tile[0].w - 1;
+                            float* outptr3 = in_tile_3.row(in_tile[0].h - 1 - i) + in_tile[0].w - 1;
+
+                            for (int j = 0; j < in_tile[0].w; j++)
+                            {
+                                float* outptr4 = in_tile_4.row(j) + i;
+                                float* outptr5 = in_tile_5.row(in_tile[0].w - 1 - j) + i;
+                                float* outptr6 = in_tile_6.row(j) + in_tile[0].h - 1 - i;
+                                float* outptr7 = in_tile_7.row(in_tile[0].w - 1 - j) + in_tile[0].h - 1 - i;
+
+                                float v = *outptr0++;
+
+                                *outptr1++ = v;
+                                *outptr2-- = v;
+                                *outptr3-- = v;
+                                *outptr4 = v;
+                                *outptr5 = v;
+                                *outptr6 = v;
+                                *outptr7 = v;
+                            }
+                        }
+                    }
+                }
+
+                // realsr
+                ncnn::Mat out_tile[8];
+                for (int ti = 0; ti < 8; ti++)
+                {
+                    ncnn::Extractor ex = net.create_extractor();
+
+                    ex.input("data", in_tile[ti]);
+
+                    ex.extract("output", out_tile[ti]);
+                }
+
+                ncnn::Mat out_alpha_tile;
+                if (channels == 4)
+                {
+                    if (scale == 1)
+                    {
+                        out_alpha_tile = in_alpha_tile;
+                    }
+                    if (scale == 4)
+                    {
+                        bicubic_4x->forward(in_alpha_tile, out_alpha_tile, opt);
+                    }
+                }
+
+                // postproc and merge alpha
+                {
+                    out.create(tile_w_nopad * scale, tile_h_nopad * scale, channels);
+                    for (int q = 0; q < 3; q++)
+                    {
+                        const ncnn::Mat out_tile_0 = out_tile[0].channel(q);
+                        const ncnn::Mat out_tile_1 = out_tile[1].channel(q);
+                        const ncnn::Mat out_tile_2 = out_tile[2].channel(q);
+                        const ncnn::Mat out_tile_3 = out_tile[3].channel(q);
+                        const ncnn::Mat out_tile_4 = out_tile[4].channel(q);
+                        const ncnn::Mat out_tile_5 = out_tile[5].channel(q);
+                        const ncnn::Mat out_tile_6 = out_tile[6].channel(q);
+                        const ncnn::Mat out_tile_7 = out_tile[7].channel(q);
+                        float* outptr = out.channel(q);
+
+                        for (int i = 0; i < out.h; i++)
+                        {
+                            const float* ptr0 = out_tile_0.row(i + prepadding * scale) + prepadding * scale;
+                            const float* ptr1 = out_tile_1.row(out_tile[0].h - 1 - i - prepadding * scale) + prepadding * scale;
+                            const float* ptr2 = out_tile_2.row(i + prepadding * scale) + out_tile[0].w - 1 - prepadding * scale;
+                            const float* ptr3 = out_tile_3.row(out_tile[0].h - 1 - i - prepadding * scale) + out_tile[0].w - 1 - prepadding * scale;
+
+                            for (int j = 0; j < out.w; j++)
+                            {
+                                const float* ptr4 = out_tile_4.row(j + prepadding * scale) + i + prepadding * scale;
+                                const float* ptr5 = out_tile_5.row(out_tile[0].w - 1 - j - prepadding * scale) + i + prepadding * scale;
+                                const float* ptr6 = out_tile_6.row(j + prepadding * scale) + out_tile[0].h - 1 - i - prepadding * scale;
+                                const float* ptr7 = out_tile_7.row(out_tile[0].w - 1 - j - prepadding * scale) + out_tile[0].h - 1 - i - prepadding * scale;
+
+                                float v = (*ptr0++ + *ptr1++ + *ptr2-- + *ptr3-- + *ptr4 + *ptr5 + *ptr6 + *ptr7) / 8;
+
+                                *outptr++ = v * 255.f + 0.5f;
+                            }
+                        }
+                    }
+
+                    if (channels == 4)
+                    {
+                        memcpy(out.channel_range(3, 1), out_alpha_tile, out_alpha_tile.total() * sizeof(float));
+                    }
+                }
+            }
+            else
+            {
+                // split alpha and preproc
+                ncnn::Mat in_tile;
+                ncnn::Mat in_alpha_tile;
+                {
+                    in_tile.create(in.w, in.h, 3);
+                    for (int q = 0; q < 3; q++)
+                    {
+                        const float* ptr = in.channel(q);
+                        float* outptr = in_tile.channel(q);
+
+                        for (int i = 0; i < in.w * in.h; i++)
+                        {
+                            *outptr++ = *ptr++ * (1 / 255.f);
+                        }
+                    }
+
+                    if (channels == 4)
+                    {
+                        in_alpha_tile = in.channel_range(3, 1).clone();
+                    }
+                }
+
+                // border padding
+                {
+                    int pad_top = std::max(prepadding - yi * TILE_SIZE_Y, 0);
+                    int pad_bottom = std::max(std::min((yi + 1) * TILE_SIZE_Y + prepadding - h, prepadding), 0);
+                    int pad_left = std::max(prepadding - xi * TILE_SIZE_X, 0);
+                    int pad_right = std::max(std::min((xi + 1) * TILE_SIZE_X + prepadding - w, prepadding), 0);
+
+                    ncnn::Mat in_tile_padded;
+                    ncnn::copy_make_border(in_tile, in_tile_padded, pad_top, pad_bottom, pad_left, pad_right, 2, 0.f, net.opt);
+                    in_tile = in_tile_padded;
+                }
+
+                // realsr
+                ncnn::Mat out_tile;
+                {
+                    ncnn::Extractor ex = net.create_extractor();
+
+                    ex.input("data", in_tile);
+
+                    ex.extract("output", out_tile);
+                }
+
+                ncnn::Mat out_alpha_tile;
+                if (channels == 4)
+                {
+                    if (scale == 1)
+                    {
+                        out_alpha_tile = in_alpha_tile;
+                    }
+                    if (scale == 4)
+                    {
+                        bicubic_4x->forward(in_alpha_tile, out_alpha_tile, opt);
+                    }
+                }
+
+                // postproc and merge alpha
+                {
+                    out.create(tile_w_nopad * scale, tile_h_nopad * scale, channels);
+                    for (int q = 0; q < 3; q++)
+                    {
+                        float* outptr = out.channel(q);
+
+                        for (int i = 0; i < out.h; i++)
+                        {
+                            const float* ptr = out_tile.channel(q).row(i + prepadding * scale) + prepadding * scale;
+
+                            for (int j = 0; j < out.w; j++)
+                            {
+                                *outptr++ = *ptr++ * 255.f + 0.5f;
+                            }
+                        }
+                    }
+
+                    if (channels == 4)
+                    {
+                        memcpy(out.channel_range(3, 1), out_alpha_tile, out_alpha_tile.total() * sizeof(float));
+                    }
+                }
+            }
+
+            {
+                if (channels == 3)
+                {
+#if _WIN32
+                    out.to_pixels((unsigned char*)outimage.data + yi * scale * TILE_SIZE_Y * w * scale * channels + xi * scale * TILE_SIZE_X * channels, ncnn::Mat::PIXEL_RGB2BGR, w * scale * channels);
+#else
+                    out.to_pixels((unsigned char*)outimage.data + yi * scale * TILE_SIZE_Y * w * scale * channels + xi * scale * TILE_SIZE_X * channels, ncnn::Mat::PIXEL_RGB, w * scale * channels);
+#endif
+                }
+                if (channels == 4)
+                {
+#if _WIN32
+                    out.to_pixels((unsigned char*)outimage.data + yi * scale * TILE_SIZE_Y * w * scale * channels + xi * scale * TILE_SIZE_X * channels, ncnn::Mat::PIXEL_RGBA2BGRA, w * scale * channels);
+#else
+                    out.to_pixels((unsigned char*)outimage.data + yi * scale * TILE_SIZE_Y * w * scale * channels + xi * scale * TILE_SIZE_X * channels, ncnn::Mat::PIXEL_RGBA, w * scale * channels);
+#endif
+                }
+            }
+        }
+    }
 
     return 0;
 }
