@@ -13,7 +13,7 @@
 class RealSR
 {
 public:
-    RealSR(int gpuid, bool tta_mode = false);
+    RealSR(int gpuid, bool tta_mode = false, int num_threads = 1);
     ~RealSR();
 
 #if _WIN32
@@ -24,6 +24,8 @@ public:
 
     int process(const ncnn::Mat& inimage, ncnn::Mat& outimage) const;
 
+    int process_cpu(const ncnn::Mat& inimage, ncnn::Mat& outimage) const;
+
 public:
     // realsr parameters
     int scale;
@@ -31,6 +33,7 @@ public:
     int prepadding;
 
 private:
+    ncnn::VulkanDevice* vkdev;
     ncnn::Net net;
     ncnn::Pipeline* realsr_preproc;
     ncnn::Pipeline* realsr_postproc;
